#include <stdarg.h>

#include "cgate.h"
#include "cgutils.h"
#include "cgutils/filesystem.h"
#include "cgutils/initemp.h"
#include "cgutils/rdtsc.h"
#ifdef _WIN32
	#include <winsock2.h>
	#include <Ws2tcpip.h>
#endif
#include "router.h"

#ifndef _WIN32
#include <signal.h>
#endif

#ifdef _WIN32
#define rmdir _rmdir
#endif
class FreePortFinder
{
public:
FreePortFinder() :
	_fd(-1)
{
#ifdef _WIN32
	_state = WSAStartup(MAKEWORD(2, 2), &_wsaData);
#endif //_WIN32
}

~FreePortFinder()
{
#ifdef _WIN32
	if (!_state)
	{
		WSACleanup();
	}
#endif //_WIN32
}

int reservePort()
{
	int port = 0;
	struct addrinfo *ptr, *addr = getAddrInfo();

	if (!addr) {
		return 0;
	}

	int fd = -1;
	for (ptr = addr; ptr; ptr = ptr->ai_next) {
		if (ptr->ai_family != AF_INET)
			continue;
		fd = (int)socket(ptr->ai_family, ptr->ai_socktype, ptr->ai_protocol);
		if (fd < 0) continue;

		int optval = 1;
		if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *)&optval, sizeof(optval))) 
		{
			closesocket(fd);
			fd = -1;
			continue;
		}
		if (!bind(fd, ptr->ai_addr, (int)ptr->ai_addrlen))
		{
			break;
		}
		fd = -1;
	}
	freeaddrinfo(addr);

	struct sockaddr_in sockAddr;
	socklen_t addrLen = sizeof(sockAddr);
	if (fd == -1)
		return 0;

	int res = getsockname(fd, (struct sockaddr *) &sockAddr,&addrLen);
	if (res == -1)
	{
		closesocket(fd);
		return 0;
	}

	if (sockAddr.sin_family == AF_INET)
		port = ntohs(sockAddr.sin_port);

	_fd = fd;
	return port;
}

void freePort()
{
	closesocket(_fd);
	_fd = -1;
}

protected:

struct addrinfo * getAddrInfo() 
{
	struct addrinfo *addr;
	struct addrinfo hints = 
	{
		AI_PASSIVE,
		AF_INET,
		SOCK_STREAM
	}; 

	int r;
	if ((r = getaddrinfo("127.0.0.1", NULL, &hints, &addr)) < 0) {
		errno = r;
		return 0;
	}
	return addr;
}

void closesocket(int fd)
{
	if (fd == -1)
	{
		return;
	}
#if !defined (_WIN32) && !defined(_WIN64)
	close(fd);
#else // !defined (_WIN32) && !defined(_WIN64)
	::closesocket(fd);
#endif // !defined (_WIN32) && !defined(_WIN64)
}

private:
#ifdef _WIN32
	WSADATA _wsaData;
#endif //_WIN32
	int _state;
	int _fd;
};

void Router::eraseTemp()
{
	if (!tempdir.length())
		return;
	remove(routerTempFileIni.c_str());
	remove(logfile.c_str());
	remove(tracefile.c_str());
	rmdir(tempdir.c_str());
	routerTempFileIni.clear();
	logfile.clear();
	tracefile.clear();
	tempdir.clear();
}

int Router::stop(bool exit)
{
	int r = proc.stop();
	if (exit)
		eraseTemp();
	return r;
}

int Router::start(const std::string& iniFile, const std::string pathExe)
{
	std::vector<std::string> v;
#ifdef _WIN32
#ifdef _WIN64
	std::string routerExe = "P2MQRouter64.exe";
#else
	std::string routerExe = "P2MQRouter.exe";
#endif
	if(!pathExe.empty())
	{
		routerExe = "/" + routerExe;
	}

	v.push_back(pathExe + routerExe);
#else
	v.push_back("P2MQRouter");
#endif
	v.push_back("/ini:" + iniFile);
	proc = Subprocess(v);
	proc.start();

	return 0;
}

int Router::start(const std::string& pathExe, TStrMap& valueMap, int& port)
{
	return start(pathExe, valueMap, port, routerIni);
}

int Router::startDirect(const std::string& pathExe, TStrMap& valueMap, int& port)
{
	cgutils::TStrStrMap routerIniDirect(routerIni);
	cgutils::TStrStrMap::iterator itr = routerIniDirect.find("ROUTEINFO");
	if (itr != routerIniDirect.end())
	{
		cgutils::TStrMap& vals(itr->second);
		cgutils::TStrMap::iterator itrDefault=vals.find("default");
		vals.erase(itrDefault);
		vals.insert(cgutils::TStrMap::value_type("direct","{routerdirect}"));
	}
	return start(pathExe, valueMap, port, routerIniDirect);
}

int Router::start( const std::string& pathExe , TStrMap& valueMap, int& port, const cgutils::TStrStrMap& iniData)
{
	FreePortFinder	freePortFinder;
	if (port != -1)
	{

		port = freePortFinder.reservePort();
		this->port = port;
	}
	else
	{
		port = this->port;
	}
	tempdir = cgutils::filesystem::mkdtemp();
	routerTempFileIni = tempdir + "/router.ini";
	logfile = tempdir + "/router.log";
	tracefile = tempdir + "/trace.ini";
	cgutils::TStrMap applyMap= create_map<std::string, std::string>("{logfile}", logfile)("{tracefile}", tracefile)("{port}", StringUtils::int2str(port));
	applyMap.insert(valueMap.begin(), valueMap.end());
	if (applyMap.find("{username}") == applyMap.end())
	{
		applyMap.insert(cgutils::TStrMap::value_type("{username}", "FORTS_HB00_rama"));//StringUtils::format("r1_%d", this->port)));
	}
	cgutils::initemp::writeMapToIni(routerTempFileIni, iniData, applyMap);
	freePortFinder.freePort();
	return start(routerTempFileIni, pathExe);
}

int Router::restore(const std::string& pathExe)
{
	return start(routerTempFileIni, pathExe);
}
	
int Router::check()
{
#ifdef _WIN32
	std::vector<HWND> windowsToClose;
	HWND windowHandle = FindWindow(NULL, NULL);
	while (windowHandle != 0)
	{
		size_t len = GetWindowTextLength(windowHandle);
		if (len)
		{
			std::string buf;
			buf.resize(len+2);
			GetWindowText(windowHandle, &buf[0], (int)len+1);
			char portStr[15];
			sprintf_s(portStr, ": %d", port);
			if (buf.find("P2MQRouter") != buf.npos && buf.find(portStr) != buf.npos )
			{
				windowsToClose.push_back(windowHandle);
			}
		}
		windowHandle = GetWindow(windowHandle, GW_HWNDNEXT);
	}
	BOOL res = 1;
	for (size_t i = 0; i < windowsToClose.size(); ++i)
	{
		res &= PostMessage(windowsToClose[i], WM_CLOSE, 0, 0);
		Sleep(1000);
	}
	return res;
#else
#endif
	return 0;
}

int Router::wait(double timeout)
{
	cg_conn_t *conn;
	std::string url = cgutils::StringUtils::format("p2lrpcq://127.0.0.1:%d;app_name=wait_for_%d", port, port);
	if (cg_conn_new(url.c_str(), &conn))
		return CG_ERR_INTERNAL;
	uint64_t deadline = cg_usec() + (uint64_t) (1000000 * timeout); //XXX: EPEL-5
	while (cg_usec() < deadline) {
		uint32_t state;
		cg_conn_getstate(conn, &state);
		if (state == CG_STATE_ACTIVE) {
			//printf("Router started in %.3f\n", timeout - (deadline - cg_usec()) / 1000000.);
			cg_conn_destroy(conn);
			return CG_ERR_OK;
		} else if (state == CG_STATE_ERROR)
			cg_conn_close(conn);
		else if (state == CG_STATE_CLOSED && cg_conn_open(conn, 0))
			cg_conn_close(conn);
		usleep(1000);
	}
	cg_conn_destroy(conn);
	return CG_ERR_TIMEOUT;
}
